import { groupBy, map, mapObjIndexed, pipe, sum } from "ramda";
import { StrictOmit } from "ts-essentials";
import { UploadStatisticsMimeRecord } from "@upload-io/upload-shared-iso/statistics/storage/dtos/shared/UploadStatisticsMimeRecord";
import { StorageStatisticsSegmentDto } from "@upload-io/upload-shared-iso/statistics/storage/dtos/shared/StorageStatisticsSegmentDto";
import { UploadStatisticsDto } from "@upload-io/upload-shared-iso/statistics/storage/dtos/shared/UploadStatisticsDto";
import { UploadStatisticsMimeSegment } from "@upload-io/upload-shared-iso/statistics/storage/dtos/shared/UploadStatisticsMimeSegment";
import { TypedFileInfo } from "@upload-io/upload-shared-iso/statistics/storage/TypedFileInfo";
import { BigSizeInBytes, EpochMillis, SizeInBytes } from "@upload-io/upload-shared-iso/common/model/Primitives";
import { NonEmpty, NonEmptyArray } from "@upload-io/upload-shared-iso/common/model/NonEmptyArray";
import { BasicFileInfo } from "@upload-io/upload-shared-iso/statistics/storage/BasicFileInfo";
import { isDefined } from "@upload-io/upload-shared-iso/common/TypeUtils";
import { IsoTimeUtils } from "@upload-io/upload-shared-iso/common/IsoTimeUtils";
import { BoxedArithmetic } from "@upload-io/upload-shared-iso/common/BoxedArithmetic";
import { BandwidthByEpochDto } from "@upload-io/upload-shared-iso/statistics/storage/dtos/shared/BandwidthByEpochDto";
import { BandwidthDto } from "@upload-io/upload-shared-iso/statistics/storage/dtos/shared/BandwidthDto";

type Predicate<T> = (item: T) => boolean;
type MimeFilters = StrictOmit<UploadStatisticsMimeRecord<Predicate<TypedFileInfo>>, "other">;
type EventsByMime = UploadStatisticsMimeRecord<TypedFileInfo[]>;
type StatisticsByMime = UploadStatisticsMimeRecord<StorageStatisticsSegmentDto>;
type EventsFunc<T> = (events: TypedFileInfo[]) => T;

const mimeFilters: MimeFilters = {
  audio: x => x.mime.startsWith("audio/"),
  image: x => x.mime.startsWith("image/"),
  video: x => x.mime.startsWith("video/"),
  document: x =>
    [
      "application/x-abiword",
      "application/msword",
      "application/vnd.ms-",
      "application/vnd.openxmlformats",
      "application/vnd.oasis.opendocument",
      "application/pdf"
    ].some(prefix => x.mime.startsWith(prefix))
};

const mimeFiltersEntries: Array<[UploadStatisticsMimeSegment, Predicate<TypedFileInfo>]> = Object.entries(
  mimeFilters
).map(([k, v]) => [k as UploadStatisticsMimeSegment, v]);

export class UploadStatisticsCalculator {
  static calculateStatistics(newEvents: NonEmptyArray<TypedFileInfo>, valueMultiplier: 1 | -1): UploadStatisticsDto {
    const groupByMime: EventsFunc<EventsByMime> = groupBy(event => {
      const match = mimeFiltersEntries.find(([_, filter]) => filter(event));
      if (match !== undefined) {
        return match[0];
      }
      return "other";
    });

    const calculateStatisticsForMime = map<EventsByMime, StatisticsByMime>(
      (events: TypedFileInfo[]) => this.calculateStatisticsSegment(NonEmpty(events), valueMultiplier) // Input is from a groupBy, so we know is non-empty.
    );

    const newStatisticsMime: Partial<StatisticsByMime> = pipe(groupByMime, calculateStatisticsForMime)(newEvents);

    const fileCreationEvents = valueMultiplier === -1 ? [] : newEvents.filter(TypedFileInfo.hasSize); // All '+1' events should include a size (we could improve types to prove this).

    return {
      all: this.calculateStatisticsSegment(newEvents, valueMultiplier),
      mime: {
        ...UploadStatisticsDto.empty.mime,
        ...newStatisticsMime
      },
      uploadBandwidth: this.calculateBandwidth(fileCreationEvents)
    };
  }

  static calculateBandwidth(bandwidthEvents: Array<BasicFileInfo<EpochMillis>>): BandwidthDto {
    const now = EpochMillis(Date.now());
    return {
      daily: this.calculateBandwidthPerEpochUnit(IsoTimeUtils.toEpochDay, 60, bandwidthEvents, now),
      monthly: this.calculateBandwidthPerEpochUnit(IsoTimeUtils.toEpochMonth, 24, bandwidthEvents, now)
    };
  }

  static calculateStatisticsSegment(
    newEvents: NonEmptyArray<BasicFileInfo>,
    valueMultiplier: 1 | -1
  ): StorageStatisticsSegmentDto {
    const timestamps = newEvents.map(x => x.created).filter(isDefined);
    return {
      firstFileCreated: timestamps.length === 0 ? undefined : EpochMillis(Math.min(...timestamps)),
      lastFileCreated: timestamps.length === 0 ? undefined : EpochMillis(Math.max(...timestamps)),
      totalFileCount: valueMultiplier * newEvents.length,
      totalFileSize: SizeInBytes(valueMultiplier * sum(newEvents.map(x => x.size))) // Can go negative (in a shard)
    };
  }

  private static calculateBandwidthPerEpochUnit(
    getEpochUnit: (time: EpochMillis) => number,
    keepLastNUnits: number,
    bandwidthEvents: Array<BasicFileInfo<EpochMillis>>,
    now: EpochMillis
  ): BandwidthByEpochDto {
    const bandwidthEventsByMonth = groupBy<BasicFileInfo<EpochMillis>>(
      x => getEpochUnit(x.created).toString(),
      bandwidthEvents
    );
    const totalSize = (events: Array<BasicFileInfo<EpochMillis>>): BigSizeInBytes =>
      BoxedArithmetic.bigSizeInBytes.sum(events.map(event => BigSizeInBytes(event.size.toString())));

    return IsoTimeUtils.trimEpochKeys(
      getEpochUnit,
      now,
      keepLastNUnits,
      mapObjIndexed(totalSize, bandwidthEventsByMonth)
    );
  }
}
